proc debugwin_create {{oldWin ""}} {
	global DeetOptActive DeetOptButtons DeetOptBalloon DeetFile DeetSource

	set w [window_create "" Debugger]
	xbpmenu_init

	set buttons [list \
		"Continue:el:cont $w"  "Stop:el:stop $w" \
		"Step:el:step $w"  "Next:el:next $w" Print:el:xprint \
		"Stack:el:xwhere $w" "Up:el:up 1 $w" "Down:el:down 1 $w" \
		Dump:el:xdump
	]

	set menu1 {
		"Open Program:target" "Open Source File:xloadfile $w"
		"Load Extension:" "File Locations:xlistfiles"
		"New Window:debugwin_create $w" "Quit:deet_close ; exit"
	}

	set menu3 [list \
		"Update on Breakpoint:DeetOptActive($w)" \
		"Show Line Numbers:DeetOptNumber($w)" \
		"Show Buttons:DeetOptButtons($w)" \
		"Value Balloons:DeetOptBalloon($w)"
	]

#	option add *Debugger*source.t.background grey80 widgetDefault
	option add *Debugger*source.t.background white widgetDefault
	window_title $w "Initializing..."
	message_create $w.message "Welcome"

	text_create $w.source
	if {[winfo depth $w] > 1} {
	   text_maketag $w.source bp yellow
	   text_maketag $w.source dbp {-background yellow -bgstipple gray25}
	} else {
	   text_maketag $w.source bp grey 
	   text_maketag $w.source dbp {-background black -foreground white \
		-bgstipple gray50}
	}
	text_maketag $w.source loc reverse
	set t [text_subwidget $w.source text]
	bind $t <ButtonPress-3> "debugwin_selline $w %x %y"
	bind $t <3> "debugwin_selline $w %x %y"
	bind $t <Double-1> "debugwin_bpset $w %x %y"

 	balloon_create $w.source "debugwin_balloon $w.source"
	optbtn_create $w.file File: "debugwin_setfile $w \[optbtn_get $w.file\]"

	for {set i 0} {$i < [llength $buttons]} {incr i} {
		set btn [split [lindex $buttons $i] :]
		lappend arg "[lindex $btn 0]:[lindex $btn 1]"
	}
	eval buttons_create $w.buttons $arg
	for {set i 0} {$i < [llength $buttons]} {incr i} {
		set btn [split [lindex $buttons $i] :]
		buttons_command $w.buttons [expr $i + 1] [lindex $btn 2]
	}
	menu_create $w.menu File Program Window Extensions Help:r
	foreach cmd $menu1 {
		menu_addcommand $w.menu 1 [lindex [split $cmd :] 0] \
			[lindex [split $cmd :] 1]
	}
	foreach cmd $buttons {
		menu_addcommand $w.menu 2 [lindex [split $cmd :] 0] \
			[lindex [split $cmd :] 2]
	}
	menu_addcommand $w.menu 2 Reset {deet_open $debug_target $debug_args}
	menu_addcommand $w.menu 2 "Invocation args" {setargs}
	menu_addcommand $w.menu 3 Breakpoints xbreakpoints 
	menu_addcommand $w.menu 3 Find... "xfind $w"
	menu_addcommand $w.menu 3 Variables xprint_create
	foreach cmd $menu3 {
		menu_addcheck $w.menu 3 [lindex [split $cmd :] 0] \
			[lindex [split $cmd :] 1]
	}
	mpack $w.menu $w.buttons:4x $w.file:Px $w.message:b $w.source:ez

	set DeetOptActive($w) 1
	set DeetOptButtons($w) 1
	set DeetOptBalloon($w) 1
	set DeetFile($w) ""
	eval optbtn_add $w.file [array names DeetSource]
	if {! [string match "" $oldWin]} {
		debugwin_setfile $w $DeetFile($oldWin)
	}
	return $w
}

proc debugwin_destroy {w} {
	global DeetOptActive DeetOptButtons DeetOptBalloon DeetFile DeetSource
	unset DeetOptActive($w)
	unset DeetOptButtons($w)
	unset DeetOptBalloon($w)
	unset DeetFile($w)
	destroy $w
}

proc debugwin_addextension {name command} {
	global DeetFile
	foreach w [array names $DeetFile] {
		menu_addcommand $w.menu 4 $name $command
	}
}

# Called to change the file in the debugging window
proc debugwin_setfile {w {file ""} {line ""}} {
	global	DeetFile DeetOptNumber breakpoint DeetLocation \
		DeetSourcePath DeetSource DeetLastMatch

	set t $w.source
	if {$DeetFile($w) != $file} {
		if {$file != ""} {
			optbtn_set $w.file $file
		} else {
			set file [optbtn_get $w.file]
		}

		window_title $w "Deet: $file" $file

		if {! [info exists DeetSource($file)]} {
			set DeetSource($file) $file
			# optbtn_add $w.file $file
		}
		if {[info exists DeetSource($file)]} {
			set f $DeetSource($file)
		} else {
			set f $file
		}
		set path $f
		if {! [sh_vnull DeetSourcePath] && ! [string match /* $f]} {
			foreach n [split $DeetSourcePath :] {
				if [file exists $n/$f] {
					set path $n/$f
					break
				}
			}
		}
		if {[catch {text_setfile $t $path}]} {
			message_set $w.message "$f, line $line"
			return
		}
		set DeetFile($w) $file
		optbtn_set $w.file $file
		foreach i [array names breakpoint] {
			if [string match $file:* $i] {
				set i [shpat_% $i {\..*}]
				set i [shpat_## $i {.*:}]
				text_tagline $t $i bp
			}
		}
		set DeetLastMatch($w) 0.0
		if {$DeetOptNumber($w)} {text_numberlines $t}
	}

	if [string match $file [lindex $DeetLocation 0]] {
		set l [lindex $DeetLocation 1]
		text_untagline $t 0 sel
		text_untagline $t 0 loc
		text_tagline $t $l loc
		text_showline $t $l.0
	}

	if {$line != ""} {
		text_untagline $t 0 sel
		text_tagline $t $line sel
		text_showline $t $line.0
	}
}

proc debugwin_balloon {w} {
	global DeetOptBalloon
	balloon_setmessage $w ""
	if {! $DeetOptBalloon([winfo parent $w])} return
	set tw [text_subwidget $w text]
	set pos "@[lindex [window_pos $tw] 0],[lindex [window_pos $tw] 1]"
	if {[text_hastag $w [$tw index $pos] sel]} {
		set t [text_get $w sel]
	} else {
        	set t [text_get $w curword]
	}
	set t [string trim $t]
	if {$t == ""} return
	if {[string match *\n* $t]} return
	if {[string match {*[=\\]*} $t]} return
        if {! [catch {set i [deet_sym -n $t]}]} {
		set dt [lindex [deet_type [lindex $i 1]] 0]
		if {$dt == "array"} {
			balloon_setmessage $w "$t = \[array\]"
		} elseif {$dt == "struct"} {
			balloon_setmessage $w "$t = \[struct\]"
		} else {
			set n [deet_getval [lindex $i 1] [lindex $i 2]]
			balloon_setmessage $w "$t = $n"
		}
        }
}

proc DeetOptNumberSet {name sub ops} {
	global DeetFile
	set file $DeetFile($sub)
	set DeetFile($sub) ""
	debugwin_setfile $sub $file
}

proc DeetOptButtonsSet {name sub ops} {
	global DeetOptButtons
	if [expr $DeetOptButtons($sub) == 0] {
		pack forget $sub.buttons
	} else {
		pack $sub.buttons -padx 4 -pady 4 -fill x -after $sub.menu
	}
}

trace var DeetOptNumber  w DeetOptNumberSet
trace var DeetOptButtons w DeetOptButtonsSet

proc debugwin_bp {mode file l c} {
	global DeetFile DeetOptActive
	foreach w [array names DeetFile] {
		switch $mode {
		   on {
			if [string match $DeetFile($w) $file] {
				text_untagline $w.source 0 sel
				text_untagline $w.source $l dbp
				text_tagline $w.source $l bp
			}
		   }
		   dis {
			if [string match $DeetFile($w) $file] {
				text_untagline $w.source 0 sel
				text_tagline $w.source $l dbp
			}
		   }
		   off {
			if [string match $DeetFile($w) $file] {
				text_untagline $w.source $l bp
			}
		   }
		   show {
			if {$DeetOptActive($w)} {
			   if {$file != ""} {
				debugwin_setfile $w $file $l
				text_untagline $w.source 0 sel
				message_set $w.message \
				   "Breakpoint in file $file, line $l"
			   } else {
				text_untagline $w.source 0 loc
				message_set $w.message "Program Finished"
			   }
			}
		   }
		}
	}
}


# called to select a line with the right mouse button
proc debugwin_selline {w x y} {
	global DeetFile

	set t $w.source
	set line [text_linenum $t $x $y]
	set b [deet_breakpoint -l $DeetFile($w) $line 0]

	if {[lindex $b 0] == ""} {return}
	text_untagline $t 0 sel
	text_tagline $t $line sel
	xbpmenu_display $w [ucvtbp [lindex $b 0]] $x $y
	if {[lindex $b 1] == ""} {return}
	foreach i $b {
	   set c [lindex $i 2]
	   [text_subwidget $t text] tag remove sel "$line.$c -1c" "$line.$c"
	}
}

# called to set/remove a breakpoint by double-clicking
proc debugwin_bpset {w x y} {
	global DeetFile breakpoint

	set t $w.source
	set line [text_linenum $t $x $y]
	set char [text_charnum $t $x $y]

	text_untagline $t 0 sel
	set bps [deet_breakpoint -l $DeetFile($w) $line 0]
	set low [lindex $bps 0]

	if [expr [llength $bps] == 0] {
		# No breakpoint on line
		message_set $w.message "No breakpoint at line $line"
		return
	} elseif {[expr [llength $bps] > 1]} {
		# More than 1 breakpoint on line
		foreach bp $bps {
			set c [lindex $low 2]
			if [expr abs([lindex $bp 2] - $char) < abs($c-$char)] { 
				set low $bp
			}
		}
	}
	
	set f [lindex $low 0]
	set l [lindex $low 1]
	set c [lindex $low 2]
	if {[info exists breakpoint($f:$l.$c)]} {
		delete $f:$l.$c
		message_set $w.message "Breakpoint removed at line $l char $c"
	} {
		b $f:$l.$c
		message_set $w.message "Breakpoint set at line $l char $c"
	}
}

proc xlistfiles_change {w} {
	global DeetSource
	set file [getfile_open "New location for $DeetSource($w)" \
		$DeetSource($w)]
	if {$file != ""} {
		set DeetSource($w) $file
	}
}

proc xlistfiles {} {
	global DeetSource
	set i 0
	set w .files
	catch {destroy $w}
	toplevel $w
	foreach f [array names DeetSource] {
		frame $w.f$i
		label $w.f$i.l -text $f
		entry $w.f$i.e -width 50 -textvar DeetSource($f)
		button $w.f$i.b -text "Change" -command "xlistfiles_change $f"
		pack $w.f$i.b $w.f$i.e -side right
		pack $w.f$i.l -side left -expand 1 -fill x
		pack $w.f$i -side top -fill x
		incr i
	}
	pack [button $w.ok -text OK -command "destroy $w"] -side bottom
}

proc xloadfile {w} {
	set file [getfile_open "Select source to open:"]
	if {$file != ""} {
		debugwin_setfile $w $file
	}
}

proc xfind {win} {
	set w [window_create]
	window_title $w "Find..."
	entry_create $w.ent "Find:"
	frame $w.f
	frame $w.f.b
	buttons_create $w.f.b.but OK:ry Close:ry
	buttons_command $w.f.b.but 1 "xfind_action $w $win"
	buttons_command $w.f.b.but 2 "destroy $w"
	checkbutton_create $w.f.b.c "Search from beginning"
	radiobox_create $w.f.rad
	radiobox_add $w.f.rad "String"
	radiobox_add $w.f.rad "Function"
	mpack $w.f:bze $w.ent:4ex $w.f.rad:lze $w.f.b:4ly $w.f.b.c:t \
		$w.f.b.but:bx
	focus $w.ent.e
}

proc xfind_action {w win} {
	global DeetLastMatch
	set str [entry_get $w.ent]
	if {[radiobox_get $w.f.rad] == "String"} {
		if {[checkbutton_get $w.f.b.c]} {
			set DeetLastMatch($win) 0.0
		}
		set t [text_subwidget $win.source text]
		set cur [$t search -count length $str "$DeetLastMatch($win)+1c"]
		if {$cur != ""} {
			text_untagline $win.source 0 sel
			$t tag add sel $cur "$cur + $length char"
			set DeetLastMatch($win) $cur
			text_showline $win.source $cur
		} else {
			set DeetLastMatch($win) 0.0
		}
	} else {
		if {! [catch {set n [lindex [deet_sym -n $str] 0]}]} {
			set arg [cvtbp $n]
			debugwin_setfile $win [lindex $arg 0] [lindex $arg 1]
		}
	}
}

proc xbpmenu_init {} {
	if {[winfo exists .bpmenu1]} return
	set m .bpmenu1
	menu $m -tearoff 0
	$m add command -label "Set Breakpoint"
	$m add command -label "Run Until Here"
	set m .bpmenu2
	menu $m -tearoff 0
	$m add command -label "Remove Breakpoint"
	$m add command -label "Disable Breakpoint"
	$m add command -label "Set Condition"
	$m add command -label "Run Until Here"
}

proc xbpmenu_display {w bp x y} {
	global breakpoint disabled
	set xpos [expr $x + [winfo rootx $w.source]]
	set ypos [expr $y + [winfo rooty $w.source]]
	if {[info exists breakpoint($bp)] || [info exists disabled($bp)]} {
		set m .bpmenu2
		$m entryconfig 0 -command "delete $bp"
		if {[info exists disabled($bp)]} {
			$m entryconfig 1 -label "Enable Breakpoint" -command \
				"enable $bp"
		} else {
			$m entryconfig 1 -label "Disable Breakpoint" -command \
				"disable $bp"
		}
		$m entryconfig 2 -command "xbreakpoints_setcond $bp"
		$m entryconfig 3 -command "deetgoto $w $bp"
	} else {
		set m .bpmenu1
		$m entryconfig 0 -command "b $bp"
		$m entryconfig 1 -command "deetgoto $w $bp"
	}
	tk_popup $m $xpos $ypos
}

proc setargs {} {
	global debug_target debug_args
	set result [dialog_entry "Command line arguments for $debug_target:" \
		"Invocation Args" $debug_args]
	if { ($result != "") && ($result != $debug_args)} {
		set debug_args $result
		deet_open $debug_target $debug_args
	}
}
